goog.provide 'app.DocumentStore'

goog.require 'app.BlockType'
goog.require 'app.document.Document'
goog.require 'goog.events.EventTarget'
goog.require 'app.document.Document'
goog.require 'app.document.Block'
goog.require 'app.document.List'
goog.require 'app.document.Image'
goog.require 'app.document.Paragraph'
goog.require 'devel.seed'

class app.DocumentStore
  ###*
    @fileoverview
    Document store provides access to document for other parts of the application.
    It has custom getters and boolean tests too.
    Speaking of CRUD, it provides CREATE and READ.
  ###

  ###*
    @const default text block
  ###
  @DefaultTextBlock: app.BlockType.Paragraph
  ###*
    @const default structured block
  ###
  @DefaultStructuredBlock: app.BlockType.Section

  ###*
    @type {app.document.Document}
  ###
  document: null

  ###*
    @type {Object} - mapping to functions for changing current type to given
    type.
  ###
  transform: null

  ###*
    @constructor
    @param {app.document.Storage} storage
    @final
  ###
  constructor: (@storage) ->

  ###*
    @param {Object} attributes
    @return {app.document.Document}
  ###
  _createDocument: (attributes) ->
    docModel = new app.document.Document attributes
    docModel.blocks = docModel.blocks.map @createBlock
    return docModel

  ###*
    @param block {app.document.Block=|Object=}
  ###
  _createTextBlockFromAny: (block={}) =>
    if (block instanceof app.document.Block)
      return block
    else
      return switch block.type
        when app.BlockType.Paragraph
          new app.document.Paragraph(block)
        when app.BlockType.List
          new app.document.List(block)
        when app.BlockType.Image
          new app.document.Image block

  _createStructuredBlockFromAny: (block) =>
    if (block instanceof app.document.StructuredBlock)
      return block
    else
      block.blocks = block.blocks.map @createBlock
      return new app.document.StructuredBlock(block)

  _onDocumentLoad: (doc) =>
    console.log("loaded document #{doc}") if app.lamebug
    @document = @_createDocument(doc)
    return doc

  ###*
    @param {string} id
  ###
  _loadFromStorage: (id) ->
    loadResult = @storage.loadDocument(id)
    loadResult.then(@_onDocumentLoad)
    return loadResult

  ###*
    @param {string} id
    @return {Promise}
    @public
  ###
  loadDocument: (id) ->
    return @_loadFromStorage(id)

  ###*
    @public
  ###
  loadAllDocs: ->
    return @storage.loadAllDocs()

  ###*
    @param {app.document.Block} needle
    @param {app.document.Block | app.document.StructuredBlock} stack
    @returns {(app.document.Block | app.document.StructuredBlock | null)} - null if not found
  ###
  _findBlockRecursive: (needle, stack) ->
    found = null
    for block in stack.blocks
      if block == needle
        found = stack
        return stack
      else if Array.isArray(block.blocks)
        found ||= @_findBlockRecursive needle, block
    return found

  ###*
    Validates if structured block can nest structured block.
    The rules are:
      - max nesting is 5 levels, where 5th is the final block
      - there can be only 3 sections chain
    @param {app.document.StructuredBlock} structure
  ###
  canNestStructured: (structure) ->
    level = 1 # document is level 1
    sectionCount = if structure.type == app.BlockType.Section then 1 else 0

    actual = structure
    while parent = @findBlockParent actual
      level++
      sectionCount++ if parent.type == app.BlockType.Section
      actual = parent

    isValid = (level < 5 && sectionCount < 3)
    return !!isValid

  ###*
    Creates new block. When no type is defined, default types are applied.
    The structured types are recognized by >blocks< property.
    @public
    @param {Object=} block - object with initial values
    @returns {app.document.Block}
  ###
  createBlock: (block={}) =>
    # set default type
    if block.blocks?
      block.type ?= app.DocumentStore.DefaultStructuredBlock
      return @_createStructuredBlockFromAny block
    else
      block.type ?= app.DocumentStore.DefaultTextBlock
      return @_createTextBlockFromAny block

  ###*
    @param {Object=} attrs
  ###
  createNewDocument: (attrs={}) ->
    attrs.title ||= 'New document title'
    @document = @_createDocument(attrs)

  createDocumentFromExample: ->
    @document = @_getSeedDocument()

  ###*
    Creates new structured block. If invalid type is given, it creates a section.
    @public
    @param {Object=} block - object with initial values
    @returns {!app.document.StructuredBlock}
  ###
  createStructuredBlock: (block={}) =>
    block.blocks ?= []
    switch block.type
      when app.BlockType.Section
        @_validateStructured new app.document.StructuredBlock(block)
      else
        throw new Error('invalid type for structured block: ')

  _validateStructured: (structured) ->
    structured.blocks = structured.blocks.map @createBlock
    return structured

  ###*
    @param block {app.document.Block}
    @returns {(app.document.Block | null)}
    TODO: repair type to app.document.StructuredBlock
  ###
  findBlockParent: (block) ->
    return @_findBlockRecursive(block, @document)

  ###*
    @param {app.BlockType} type
    @return {app.document.Block} with given type if exists or null
  ###
  findFirstOfType: (type) ->
    isGivenType = (block) -> block.type == type
    return @_findBlockIf(isGivenType, @document)

  ###*
    @param {app.BlockType} type
    @param {app.document.Block=} structure
    @return {Array.<app.document.Block>}
  ###
  findOfType: (type, structure=@document) ->
    foundBlocks = []
    for b in structure.blocks
      if b.type == type
        foundBlocks.push b
      if b.blocks?
        foundBlocks = foundBlocks.concat @findOfType type, b

    return foundBlocks

  ###*
    TODO: repair type to app.document.StructuredBlock
    @param {function(app.document.Block):boolean} testFunc - to test blocks
    @param {app.document.Block} block - where to look
    @return {app.document.Block} - any block that pass testFunc or null
  ###
  _findBlockIf: (testFunc, block) =>
    if testFunc(block)
      return block
    else if (block.blocks?)
      for nestedBlock in block.blocks
        found = @_findBlockIf(testFunc, nestedBlock)
        return found if found?
    return null

  ###*
    @return {Array.<string>} all types of element blocks
  ###
  getElementBlockTypes: ->
    return app.BlockType.getAll()

  ###*
    @return {app.document.Document}
  ###
  getDocument: ->
    return @document

  ###*
    @return {app.document.Document}
  ###
  _getSeedDocument: ->
    return @_createDocument devel.seed
