goog.provide 'app.EditorStore'

goog.require 'app.AppState'
goog.require 'app.BlockType'
goog.require 'app.document.Document'
goog.require 'app.document.Paragraph'
goog.require 'app.document.List'
goog.require 'goog.events.EventTarget'

class app.EditorStore extends goog.events.EventTarget
  ###*
    @fileoverview
    Editor store access document from document store.
    Any change (UPDATE, DELETE) made upon a document is done in this store.
    Document component listens on the store's changes.
  ###

  ###*
    @enum {string}
  ###
  @EventType: {
    Change: 'document-change'
    AppStateChange: 'AppState-change'
  }

  ###*
    @type {Object} - mapping to functions for changing current type to given
    type.
  ###
  transform: null

  ###*
    @type {app.document.Document} - is null if not loaded yet
  ###
  doc: null

  ###*
    @type {app.AppState}
  ###
  appState: null

  ###*
    @constructor
    @extends {goog.events.EventTarget}
    @param {app.document.Storage} storage
    @param {app.DocumentStore} docStore
    @param {app.editor.TransformType} transform
    @param {app.StateStorage} stateStorage
    @final
  ###
  constructor: (@storage, @docStore, @transform, @stateStorage) ->
    @appState = new app.AppState(@stateStorage.load())
    super()

  ###*
    Saves current state of app. Saves to localStorage and dispatch event for
    TOP level component {app.editor.Document}.
    @private
  ###
  _notify: ->
    @dispatchEvent app.EditorStore.EventType.Change
    @_saveOnly()

  _notifyAppState: ->
    @dispatchEvent app.EditorStore.EventType.AppStateChange
    @_persistAppState()

  ###*
    Saves as in _notify, but doesnt inform about change. The component owner of the
    data must handle this. Consult it with React documentation.
  ###
  _saveOnly: ->
    saveResult = @storage.saveDocument(@doc)
    unless saveResult == null
      saveResult.then(
        @onDocumentSaved
        app.reportError
      )

  _persistAppState: ->
    @stateStorage.save({
      editingDocumentId: @appState.editingDocumentId
      shouldActivateBlockId: @appState.shouldActivateBlockId
    })

  activateBlockIfNeeded: (block) ->
    @setAppState({shouldActivateBlockId: block._id})

  ###*
    @param {app.document.Block} toBlock
    @param {boolean} up
  ###
  _activate_vertically: (toBlock, up) ->
    indexChange = if up then -1 else 1
    parent = @docStore.findBlockParent(toBlock)
    app.reportError('should have parent') unless parent?
    index = parent.blocks.indexOf(toBlock)
    return if index == -1
    maybeBlock = parent.blocks[index+indexChange]
    if maybeBlock?
      @setAppState({shouldActivateBlockId: maybeBlock._id})

  activateDownTo: (toBlock) ->
    @_activate_vertically(toBlock, false)

  activateInsideTo: (toBlock) ->
    return unless toBlock.type == app.BlockType.Section
    maybeBlock = toBlock.blocks[0]
    if maybeBlock?
      @setAppState({shouldActivateBlockId: maybeBlock._id})

  activateOutsideTo: (toBlock) ->
    parent = @docStore.findBlockParent(toBlock)
    if parent? && parent.type != app.BlockType.Document
      @setAppState({shouldActivateBlockId: parent._id})

  activateUpTo: (toBlock) ->
    @_activate_vertically(toBlock, true)

  ###*
    @param {app.document.Block} afterBlock
    @param {app.BlockType} type
  ###
  addBlockAfter: (afterBlock, type) ->
    parent = @docStore.findBlockParent(afterBlock)
    if parent?
      newBlock = (
        if type == app.BlockType.Section
          @docStore.createStructuredBlock({type})
        else
          @docStore.createBlock({type})
      )
      index = parent.blocks.indexOf(afterBlock)
      parent.blocks.splice(index+1, 0, newBlock)
      @_notify()

  addImageData: (imgBlock, data) ->
    imgBlock.imageData = data
    @_notify()

  ###*
    @param {app.document.StructuredBlock} structure - to append
    @param {app.BlockType} type
  ###
  appendStructure: (structure, type) ->
    section = @docStore.createStructuredBlock({type})
    structure.blocks.push section
    @_notify()

  ###*
    @param structure {app.document.StructuredBlock}
    @param type {?app.BlockType} - type of block to append
  ###
  appendTextBlock: (structure, type) ->
    structure.blocks.push @docStore.createBlock({type})
    @_notify()

  ###*
    @param {app.document.Block} block
    @param {string} title
  ###
  changeTitle: (block, title) ->
    block.title = title
    @_notify()

  ###*
    @param {app.document.Block} block
    @param {app.BlockType} newType
  ###
  changeType: (block, newType) ->
    newBlock = @transform.transformTo block, newType
    @replaceBlock(block, newBlock)
    #console.log 'newBlock', newBlock
    @_notify()

  ###*
    @param {app.document.Block} block
    @param {string} newComment
  ###
  changeComment: (block, newComment) ->
    block.comment = newComment
    @_saveOnly()

  ###*
    @param {app.document.Block} block
    @param {string} newNote
  ###
  changeNote: (block, newNote) ->
    block.authorNote = newNote
    @_saveOnly()


  ###*
    @param block {app.document.Section}
    @param changeObject {app.document.Section}
  ###
  changeSection: (block, changeObject) ->
    goog.mixin block, changeObject
    @_notify()

  ###*
    @param {app.document.Block} paragraph
    @param {string} text
  ###
  changeParagraphContent: (paragraph, text) =>
    paragraph.text = text
    @_notify()

  ###*
    @param {app.document.List} list
    @param {Array.<string>} items
  ###
  changeListContent: (list, items) =>
    list.items = items
    @_notify()

  changeListType: (list, type) ->
    list.isOrdered = if type == 'ordered' then true else false
    @_notify()

  deleteBlock: (block) =>
    parent = @docStore.findBlockParent(block)
    if (parent)
      #console.log 'removing', block
      #console.log 'removing', parent
      idx = parent.blocks.indexOf(block)
      parent.blocks.splice(idx, 1)
      @_notify()
    else
      console.warn 'parent not found'

  ###*
    @public
  ###
  deleteDocument: (doc) ->
    @storage.deleteDocument doc

  forceBlockActivation: (block) ->
    @setAppState({shouldActivateBlockId: block._id, activeBlock: null})
    # TODO should be ok to remove:
#    @_notify()

  ###*
    @public
  ###
  forceDocumentSave: ->
    @_saveOnly()

    ###*
      @public
    ###
  getDocument: ->
    return @doc

  moveUP: (block) ->
    @_moveVertically(true, block)
    @forceBlockActivation(block)

  moveDOWN: (block) ->
    @_moveVertically(false, block)
    @forceBlockActivation(block)

  moveLEFT: (block) ->
    @_moveHorizontally(true, block)
    @forceBlockActivation(block)

  moveRIGHT: (block) ->
    @_moveHorizontally(false, block)
    @forceBlockActivation(block)

  ###*
    When moving outside structure its place before the structure.
    When moving to a structure, its placed to the following one.
    @param {boolean} isMovingLeft
    @param {app.document.Block} block
  ###
  _moveHorizontally: (isMovingLeft, block) ->
    parent = @docStore.findBlockParent block
    return unless parent? # just for document

    blockIndex = parent.blocks.indexOf block
    if isMovingLeft
      grandparent = @docStore.findBlockParent parent
      return unless grandparent? # moving left when at TOP
      parent.blocks.splice blockIndex, 1
      parentIndex = grandparent.blocks.indexOf parent
      grandparent.blocks.splice parentIndex, 0, block
    else
      maybeSection = parent.blocks[blockIndex+1]
      return unless maybeSection? # moving right when at BOTTOM
      if maybeSection.type == app.BlockType.Section
        parent.blocks.splice blockIndex, 1
        maybeSection.blocks.splice 0, 0, block

  ###*
    @param {boolean} isMovingUp
    @param {app.document.Block} block
  ###
  _moveVertically: (isMovingUp, block) ->
    parent = @docStore.findBlockParent block
    blockIndex = parent.blocks.indexOf block
    return false if isMovingUp && blockIndex == 0
    return false if !isMovingUp && blockIndex == (parent.blocks.length-1)

    if parent?
      indexChange = if isMovingUp then -1 else 1
      parent.blocks.splice blockIndex, 1
      parent.blocks.splice (blockIndex + indexChange), 0, block

  onDocumentSaved: (info) =>
    @doc._rev = info.rev
    @storage.updatedRev()

  ###*
    @param {app.document.Block} oldBlock
    @param {app.document.Block} newBlock
  ###
  replaceBlock: (oldBlock, newBlock) ->
    parent = @docStore.findBlockParent(oldBlock)
    idx = parent.blocks.indexOf(oldBlock)
    parent.blocks.splice(idx, 1, newBlock)

  ###*
    @param {app.document.Document} documentObject
  ###
  setDocument: (documentObject) =>
    @doc = documentObject

  ###*
    @param {Object} changeObject
  ###
  setAppState: (changeObject) ->
    goog.mixin @appState, changeObject
    @_notifyAppState()